"""
    Security audit utility.
"""
import time

from AccessControl.Permission import Permission
from AccessControl.SecurityInfo import ClassSecurityInfo
from Acquisition import aq_base
from DateTime import DateTime
from Globals import InitializeClass
from OFS.PropertyManager import PropertyManager
from OFS.SimpleItem import SimpleItem
from Products.PageTemplates.PageTemplateFile import PageTemplateFile

class SecurityTweak(SimpleItem):
    """ Infoabout one or more security tweaks applied to an object in the ZODB.
    
    The captured tweaks include:

      - Local role assigments

      - Proxy roles.

      - Overridden permission mappings.

      - Executable ownership which differes from the ownership of
        the 'aq_parent'.

      - 'Owner' local roles which don't include the executable owner
        (leave out 'System Processes' for unowned).
    """
    __allow_access_to_unprotected_subobjects__ = 1
    local_roles = owners_by_role = proxy_roles = ()
    permission_overrides = {}
    new_owner = owner_mismatch = None
    capture_error = None
   
    def __init__( self, obj ):

        self.path = obj.getPhysicalPath()
        self.meta_type = obj.meta_type

        # Add exception handling per Wolfgang Strobl's suggestion.
        try:
            # Find local roles other than 'Owner'.
            self.local_roles = []
            self.owners_by_role = []
            for k, v in obj.get_local_roles():
                v = list( v )
                if 'Owner' in v:
                    self.owners_by_role.append( k )
                try:
                    v.remove( 'Owner' )
                except ValueError:
                    pass
                if v:
                    self.local_roles.append( ( k, tuple( v ) ) )

            # Find proxy roles, if any.
            self.proxy_roles = getattr( obj, '_proxy_roles', None )

            # Find overridden (non-acquired) permissions.
            po = self.permission_overrides = {}
            for p in obj.ac_inherited_permissions(1):
                name, value = p[:2]
                p = Permission( name, value, obj )
                roles = p.getRoles()
                if type( roles ) == type( () ):
                    po[ name ] = roles

            # Find "responsible" owner, compare with parent.
            owner = obj.getOwner( 1 )
            parent = getattr( obj, 'aq_parent', None )
            try:
                powner = parent is not None and parent.getOwner( 1 ) or None
            except AttributeError:
                powner = None
            self.new_owner = owner != powner and owner or None
            if owner:
                if owner[1] not in self.owners_by_role:
                    self.owner_mismatch = ( owner[1], self.owners_by_role )
                else:
                    self.owner_mismatch = None
            else:
                self.owner_mismatch = (
                    self.owners_by_role != ['System Processes']
                      and ( None, self.owners_by_role ) or None )
        except:
            import sys, traceback
            try:
                etype, value, tb = sys.exc_info()
                self.capture_error = traceback.format_exception( etype, value
                                                               , tb )
            finally:
                etype = value = tb = None

    def __nonzero__( self ):
        # Do we have anything interesting to say?
        return ( self.local_roles
              or self.proxy_roles
              or self.permission_overrides
              or self.new_owner
              or self.owner_mismatch
              or self.capture_error
               ) and 1 or 0


class SecurityAudit( SimpleItem, PropertyManager ):
    """ Snapshot information about the state of the ZODB w.r.t. security.
    """
    meta_type = 'Security Audit'
    security = ClassSecurityInfo()

    def __init__( self, context ):

        path = ['']
        path.extend( context.getPhysicalPath() )
        self.context_path = '/'.join( path )
        self.created = DateTime()

        tweaks = []

        for p, obj in context.ZopeFind( context, search_sub=1 ):
            tweak = SecurityTweak( obj )
            if tweak:
                tweaks.append( tweak )

        self.tweaks = tuple( tweaks )

    _properties = ( { 'id' : 'context_path', 'type' : 'string', 'mode' : 'r' },
                    { 'id' : 'created', 'type' : 'date', 'mode' : 'r' },
                  )

    manage_options = ( { 'label' : 'Properties'
                       , 'action' : 'manage_propertiesForm'
                       },
                       { 'label' : 'Audit'
                       , 'action' : 'index_html'
                       },
                     )

    index_html = PageTemplateFile( 'zmi/audit.zpt', globals() )

InitializeClass(SecurityAudit)

def manage_addSecurityAudit(dispatcher, REQUEST=None):
    """ Add a SecurityAudit to dispatcher.
    """
    timestamp = time.gmtime()
    fmt = 'secaudit-%4d%02d%02d%02d%02d%02d'
    id =  fmt % timestamp[:6]
    # Use the FactoryDispatcher's real context
    context = getattr( aq_base( dispatcher ), '_d', dispatcher )
    audit = SecurityAudit( context )
    audit._setId( id )
    dispatcher._setObject( id, audit )

    if REQUEST is not None:
        REQUEST['RESPONSE'].redirect( '%s/manage_main'
                                        % dispatcher.absolute_url() )
